import math
import os
import logging
import subprocess
import json

from pathlib import Path

import numpy as np
from tondortools.tool import read_raster_info, save_raster, mosaic_tifs
from multiprocessing import Pool

import numpy
from scipy.interpolate import Rbf
from osgeo.gdalconst import GA_ReadOnly, GDT_Float32, GDT_Int16, GDT_Byte
try:
    from osgeo import gdal
except:
    import gdal

try:
    from src.tondor.util.tool import merge_tiles, setup_tiles, raster_bounds_to_polygon, create_template_raster, read_raster_info
    from src.tondor.util.geo import BoundingBox
except:
    from tondor.util.tool import merge_tiles, setup_tiles, raster_bounds_to_polygon, create_template_raster, read_raster_info
    from tondor.util.geo import BoundingBox

from scipy.interpolate import interp1d, CubicSpline
NODATA_VALUE = -99999
MAX_TILE_SIZE = 3000
CPU_COUNT = 7

log = logging.getLogger(__name__)


def append_to_dict(dictionary, key, value):
    if key in dictionary:
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]


def generate_interpolated_multiband_tile(input_data):
    tile_info = input_data[0]

    tile_multiband_interpolated_composite = Path(tile_info["tile_multiband_interpolated_composite"])
    if not tile_multiband_interpolated_composite.exists():
        (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(
            str(tile_info['tile_multiband_composite']))
        input_rasters_bands = tile_info['n_band']

        src_ds = gdal.Open(str(tile_info['tile_multiband_composite']))
        src_ds_array = src_ds.ReadAsArray().astype(float)
        src_ds_array_shape = src_ds_array.shape
        nodata_value = src_ds.GetRasterBand(1).GetNoDataValue()

        if tile_info['tile_mask'] is not None and tile_info['tile_mask'] != "None":
            src_mask_ds = gdal.Open(str(tile_info['tile_mask']))
            src_mask_ds_array = src_mask_ds.ReadAsArray().astype(float)
            src_mask_ds_array_1d = src_mask_ds_array.ravel()
            src_mask_ds_array_band_zero_indices = numpy.where(src_mask_ds_array_1d == 0)[0]
            
        if len(src_ds_array_shape) == 3:
            array_bands = src_ds_array_shape[0]
            array_rows = src_ds_array_shape[1]
            array_cols = src_ds_array_shape[2]
        elif len(src_ds_array_shape) == 2:
            array_rows = src_ds_array_shape[0]
            array_cols = src_ds_array_shape[1]

        n_timeperiods = int(n_bands / input_rasters_bands)
        nodata_mask = numpy.zeros((n_timeperiods, ysize, xsize))

        for raster_band in range(int(input_rasters_bands)):
            zero_indices_master_complete = set
            for time_epoch in numpy.arange(n_timeperiods):
                array_band_number = raster_band + (input_rasters_bands * time_epoch)
                src_ds_array_band = src_ds_array[array_band_number, :, :]
                nodata_mask[time_epoch, :, :] += src_ds_array_band == nodata_value
                src_ds_array_1d = src_ds_array_band.ravel()
                src_ds_array_band_zero_indices = numpy.where(src_ds_array_1d == nodata_value)[0]
                zero_indices_master_complete = zero_indices_master_complete.union(set(src_ds_array_band_zero_indices))

            if tile_info['tile_mask'] is not None and tile_info['tile_mask'] != "None":
                zero_indices_master = zero_indices_master_complete.difference(src_mask_ds_array_band_zero_indices)
            else:
                zero_indices_master = zero_indices_master_complete
                

            zero_indices_master_list = list(zero_indices_master)
            if len(zero_indices_master_list) == 0 : continue

            band_allmonth_1darray_list = []

            for time_epoch in numpy.arange(n_timeperiods):
                array_band_number = raster_band + (input_rasters_bands * time_epoch)
                src_ds_array_band = src_ds_array[array_band_number, :, :]
                src_ds_array_1d = src_ds_array_band.ravel()
                band_allmonth_1darray_list.append(src_ds_array_1d[zero_indices_master_list])

            band_allmonth_1darray = numpy.stack(band_allmonth_1darray_list, axis=0)
            band_allmonth_1darray[band_allmonth_1darray == nodata_value] = NODATA_VALUE
            numpy.apply_along_axis(custom_interpolate_data, 0, band_allmonth_1darray)
            band_allmonth_1darray[band_allmonth_1darray == NODATA_VALUE] = nodata_value

            for time_epoch in numpy.arange(n_timeperiods):
                array_band_number = raster_band + (input_rasters_bands * time_epoch)
                src_ds_array_band = src_ds_array[array_band_number, :, :]
                src_ds_array_1d = src_ds_array_band.ravel()
                src_ds_array_1d[zero_indices_master_list] = band_allmonth_1darray[time_epoch, :]
                src_ds_array[array_band_number, :, :] = src_ds_array_1d.reshape(array_rows, array_cols)

        ulx = tile_info["ulx"]
        uly = tile_info["uly"]
        src_ds_array[src_ds_array == NODATA_VALUE] = nodata_value
        #nodata_mask[nodata_mask > 1] = 1
        save_raster(src_ds_array, tile_info["tile_multiband_interpolated_composite"], "GTiff", int(epsg), ulx, uly,
                    pixel_width, data_type,
                    nodata_value=nodata_value)
        save_raster(nodata_mask, tile_info["tile_multiband_mask"], "GTiff", int(epsg), ulx, uly,
                    pixel_width, GDT_Byte,
                    nodata_value=None)
        log.debug(f"- done interpolating {tile_info['row']}, {tile_info['col']} -")

def generate_multiband_tile(input_data):

    src_raster_filepaths = input_data[0]
    tile_info = input_data[1]


    multiband_raster_filepath = Path(tile_info['tile_multiband_composite'])
    first_raster = src_raster_filepaths[0]
    raster_ds = gdal.Open(str(first_raster))
    nodata_value = raster_ds.GetRasterBand(1).GetNoDataValue()

    tile_composite_filepaths = []
    for composite_filepath in src_raster_filepaths:
        tile_composite_filepaths.append(Path(tile_info['tile_folder']).joinpath(Path(composite_filepath).name))

    if not Path(multiband_raster_filepath).exists():
        args = ["gdal_merge.py",
                "-separate",
                "-o", str(multiband_raster_filepath),
                "-of", "GTiff",
                "-co", "compress=DEFLATE",
                "-a_nodata", f"{nodata_value}",
                "-ul_lr", str(tile_info["ulx"]), str(tile_info["uly"]), str(tile_info["lrx"]), str(tile_info["lry"])]
        args = args + [str(filepath) for filepath in tile_composite_filepaths]
        cmd_output = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        log.debug("exit code {} --> {}".format(cmd_output.returncode, args))

def mosaic_tile_tifs(tile_infos, mosaic_filepath, work_dir, method='average'):
    raster_list = []
    for tile_info in tile_infos:
        if Path(tile_info["tile_multiband_interpolated_composite"]).exists():
            raster_list.append(tile_info["tile_multiband_interpolated_composite"])


    src_ds = gdal.Open(str(raster_list[0]))
    nodata_value = src_ds.GetRasterBand(1).GetNoDataValue()

    mosaic_filename = Path(mosaic_filepath).name
    merge_tiles(raster_list, mosaic_filename, work_dir, method)

def generate_subtile_parallel(input_data):
    first_file = input_data[0]
    second_file = input_data[1]
    mosaic_filepath = input_data[2]
    no_data_value = int(input_data[3])
    mosaic_tifs([first_file, second_file], mosaic_filepath, no_data_value)

def recursive_mosaic(file_list, output_file, work_dir, nodata_value, process_cpu_count = CPU_COUNT):

    work_dir_interpolation = work_dir.joinpath('submosaic')
    os.makedirs(work_dir_interpolation, exist_ok=True)

    """Mosaic files recursively two at a time."""
    if len(file_list) == 1:
        os.rename(file_list[0], output_file)
        return

    temp_outputs = []

    subtasks = []
    # Pair up files and mosaic each pair
    for i in range(0, len(file_list), 2):
        if i + 1 < len(file_list):  # Ensure there's a pair
            temp_output = Path(work_dir_interpolation).joinpath(f"temp_mosaic_{i}.tif")
            if not temp_output.exists():
                args = [file_list[i], file_list[i+1], temp_output, nodata_value]
                subtasks.append(args)
            temp_outputs.append(temp_output)
        else:
            temp_outputs.append(file_list[i])
    if len(subtasks) > 0:
        p = Pool(process_cpu_count)
        p.map(generate_subtile_parallel, tuple(subtasks))

    # Recursively mosaic the temp outputs
    recursive_mosaic(temp_outputs, output_file, work_dir_interpolation, nodata_value, process_cpu_count)

    # Cleanup temporary files
    for temp in temp_outputs:
        os.remove(temp)


def do_recursive_tile_mosaic(tile_infos, mosaic_filepath, work_dir):
    raster_list = []
    for tile_info in tile_infos:
        if Path(tile_info["tile_multiband_interpolated_composite"]).exists():
            raster_list.append(tile_info["tile_multiband_interpolated_composite"])

    src_ds = gdal.Open(str(raster_list[0]))
    nodata_value = src_ds.GetRasterBand(1).GetNoDataValue()

    recursive_mosaic(raster_list, mosaic_filepath, work_dir, nodata_value)


def create_multiband_after_tiling(composite_filepaths, tile_infos, process_cpu_count = CPU_COUNT):

    composite_filepaths = [str(composite_filepath) for composite_filepath in composite_filepaths]

    subtasks = []
    for tile_info in tile_infos:
        tile_multiband_path = Path(tile_info["tile_folder"]).joinpath(tile_info["tile_multiband_composite"])
        if tile_multiband_path.exists(): continue
        args = [composite_filepaths, tile_info]
        subtasks.append(args)

    if len(subtasks) > 0:
        p = Pool(process_cpu_count)
        p.map(generate_multiband_tile, tuple(subtasks))
        p.close()  # Close the pool
        p.join()

def tiled_stats(composite_filepaths, work_dir, stat_list=['min', 'mean', 'max', 'range','std_dev'], tile_size = None):
    if tile_size is None:
        max_size_tile = MAX_TILE_SIZE
    else:
        max_size_tile = tile_size
    (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(
        composite_filepaths[0])
    tile_infos = setup_tiles(xmin, ymax, xsize, ysize, pixel_width, n_bands, max_size_tile, work_dir)

    stat_tiled_dir = work_dir.joinpath('stat')
    stat_tiled_dir.mkdir(parents=True, exist_ok=True)
    create_tilecomposites(composite_filepaths, tile_infos, process_cpu_count=CPU_COUNT)

    create_multiband_after_tiling(composite_filepaths, tile_infos)

    stat_tiled_dict = {}
    for stat_list_item in stat_list:
        stat_tiled_dict[stat_list_item] = []

    for tile_info in tile_infos:
        log.debug(f"-- tile {tile_info['row']} {tile_info['col']} --")
        tile_workdir = Path(tile_info["tile_folder"])
        tile_multiband_raster = tile_info["tile_multiband_composite"]
        ulx = tile_info["ulx"]
        uly = tile_info["uly"]

        src_ds = gdal.Open(str(tile_multiband_raster), GA_ReadOnly)
        imagery = numpy.zeros((src_ds.RasterCount, tile_info["height"], tile_info["width"]), dtype="float")
        for band_index in range(src_ds.RasterCount):
            band_data = src_ds.GetRasterBand(band_index + 1).ReadAsArray().astype("float")
            nodata_value = src_ds.GetRasterBand(1).GetNoDataValue()
            band_data[band_data == nodata_value] = numpy.nan
            imagery[band_index,:, : ] = band_data

        log.debug(f"collecting stats for {tile_info['row']}_{tile_info['col']}")
        for stat_list_item in stat_list:
            
            if stat_list_item == 'min': 
                ndvi_min_filepath  = tile_workdir.joinpath(f"min_{tile_info['row']}_{tile_info['col']}.tif")
                ndvi_min = numpy.nanmin(imagery, axis=0)
                ndvi_min[ndvi_min == numpy.nan] == NODATA_VALUE
                save_raster(ndvi_min, ndvi_min_filepath, "GTiff", int(epsg), ulx, uly, pixel_width, GDT_Float32,
                            nodata_value=NODATA_VALUE)
                del ndvi_min
                stat_tiled_dict['min'].append(ndvi_min_filepath)
    
            if stat_list_item == 'mean': 
                ndvi_mean_filepath  = tile_workdir.joinpath(f"mean_{tile_info['row']}_{tile_info['col']}.tif")
                ndvi_mean = numpy.nanmean(imagery, axis=0)
                ndvi_mean[ndvi_mean == numpy.nan] == NODATA_VALUE
                save_raster(ndvi_mean, ndvi_mean_filepath, "GTiff", int(epsg), ulx, uly, pixel_width, GDT_Float32,
                            nodata_value=NODATA_VALUE)
                del ndvi_mean
                stat_tiled_dict['mean'].append(ndvi_mean_filepath)
    
            if stat_list_item == 'max':
                ndvi_max_filepath  = tile_workdir.joinpath(f"max_{tile_info['row']}_{tile_info['col']}.tif")
                ndvi_max = numpy.nanmax(imagery, axis=0)
                ndvi_max[ndvi_max == numpy.nan] == NODATA_VALUE
                save_raster(ndvi_max, ndvi_max_filepath, "GTiff", int(epsg), ulx, uly, pixel_width, GDT_Float32,
                            nodata_value=NODATA_VALUE)
                del ndvi_max
                stat_tiled_dict['max'].append(ndvi_max_filepath)
    
            if stat_list_item == 'range':
                ndvi_range_filepath = tile_workdir.joinpath(f"range_{tile_info['row']}_{tile_info['col']}.tif")
                ndvi_max = numpy.nanmax(imagery, axis=0)
                ndvi_min = numpy.nanmin(imagery, axis=0)
                ndvi_range = ndvi_max - ndvi_min
                ndvi_range[ndvi_range == numpy.nan] == NODATA_VALUE
                save_raster(ndvi_range, ndvi_range_filepath, "GTiff", int(epsg), ulx, uly, pixel_width, GDT_Float32,
                            nodata_value=NODATA_VALUE)
                del ndvi_range
                stat_tiled_dict['range'].append(ndvi_range_filepath)
    
            if stat_list_item == 'std_dev':
                ndvi_std_filepath = tile_workdir.joinpath(f"std_{tile_info['row']}_{tile_info['col']}.tif")
                ndvi_stddev = numpy.nanstd(imagery, axis=0)
                ndvi_stddev[ndvi_stddev == numpy.nan] == NODATA_VALUE
                save_raster(ndvi_stddev, ndvi_std_filepath, "GTiff", int(epsg), ulx, uly, pixel_width, GDT_Float32,
                            nodata_value=NODATA_VALUE)
                del ndvi_stddev
                stat_tiled_dict['std_dev'].append(ndvi_std_filepath)

        del imagery

    log.debug(f"mosaicing stat rasters")
    stat_composite_list = []
    for dict_key, raster_list in stat_tiled_dict.items():
        mosaic_path = work_dir.joinpath(f"{dict_key}.tif")
        mosaic_tifs(raster_list, mosaic_path, no_data=NODATA_VALUE)
        stat_composite_list.append(mosaic_path)
    return stat_composite_list

def custom_interpolate_data(data_array):
    try:
        """
        Interpolates the given data using scipy's interp1d and CubicSpline.
        If the missing data is at the first or last position, interp1d is used.
        If the missing data is in between, CubicSpline/RBF interpolation is used.
        """
        data_array[data_array == NODATA_VALUE] = numpy.nan
        # Indices for valid and missing data points
        valid_indices = numpy.where(~numpy.isnan(data_array))[0]
        missing_indices = numpy.where(numpy.isnan(data_array))[0]

        if len(valid_indices) == 1:
            # Finding the non-NaN value
            non_nan_value = data_array[~np.isnan(data_array)][0]
            # Replicating this value across the entire array
            data_array[:] = non_nan_value
        elif len(valid_indices) == 0:
            data_array
        else:
            # Create a copy of the data array for interpolation
            interpolated_data = numpy.copy(data_array)

            # Handle missing data at the beginning or end
            if 0 in missing_indices or (len(data_array) - 1) in missing_indices:
                # Interpolate using linear interpolation
                linear_interp = interp1d(valid_indices, data_array[valid_indices],
                                        kind='linear', fill_value="extrapolate")
                for index in missing_indices:
                    interpolated_data[index] = linear_interp(index)

            if 0 in missing_indices:
                data_array[0] = interpolated_data[0]
            if (len(data_array) - 1) in missing_indices:
                data_array[(len(data_array) - 1)] = interpolated_data[(len(data_array) - 1)]

            # Handle missing data in the middle
            # Create an array of x-coordinates for non-NaN values

            x = numpy.arange(len(data_array))
            y = data_array

            # Remove NaN values for interpolation
            valid_indices_rbf = ~numpy.isnan(y)
            nonvalid_indices_rbf = numpy.isnan(y)
            x_valid = x[valid_indices_rbf]
            y_valid = y[valid_indices_rbf]

            # Perform cubic interpolation
            #cubic_spline = CubicSpline(x_valid, y_valid, bc_type='natural')
            sm = 0.2
            rbf = Rbf(x_valid, y_valid, smooth=sm)

            # Evaluate the interpolated values for all x-coordinates
            #data_pred = cubic_spline(x)
            data_pred = rbf(x)
            data_array[nonvalid_indices_rbf] = data_pred[nonvalid_indices_rbf]
    except Exception as e:
            raise  Exception(f"An error occurred: {e} {data_array} {valid_indices}")

def FitSplineRBF_array(row_array):

    if numpy.sum(row_array != NODATA_VALUE) < 3:
        return

    w = numpy.asarray(row_array == NODATA_VALUE)
    heights = numpy.linspace(1, len(row_array), len(row_array))

    x_col = heights[~w]
    y_col = row_array[~w]
    sm = 0.25
    rbf = Rbf(x_col, y_col, smooth=sm)

    x_excluded = heights[w]
    y_pred = rbf(x_excluded)

    row_array[w] = y_pred


def cut_composite_totile_function(input_data):
    src_filepath = input_data[0]
    tile_extent_path_dict_json = input_data[1]

    tile_infos = json.loads(tile_extent_path_dict_json)

    for tile_name, tile_info in tile_infos.items():
        tile_composite_filepath = Path(tile_info['tile_folder']).joinpath(Path(src_filepath).name)
        if tile_composite_filepath.exists(): continue
        cmd_gdal = ["gdal_translate",
                    "-of", "GTiff",
                    "-co", "COMPRESS=DEFLATE",
                    "-co", "BIGTIFF=YES",
                    "-co", "TILED=YES",
                    "-eco", "-projwin",
                    "{}".format(tile_info['ulx']), "{}".format(tile_info['uly']),
                    "{}".format(tile_info['lrx']), "{}".format(tile_info['lry']),
                    str(src_filepath), str(tile_composite_filepath)]
        cmd_output = subprocess.run(cmd_gdal, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        log.debug(f"exit code {cmd_output.returncode} --> {cmd_gdal}")

        if int(cmd_output.returncode) != 0:
            log.error(f"{cmd_output}")
            if "falls completely outside raster extent" in str(cmd_output.stderr):
                shapely_bbox = raster_bounds_to_polygon(src_filepath)
                (xmin, ymax, RasterXSize, RasterYSize, pixel_width, projection, epsg, datatype, n_bands, imagery_extent_box) = read_raster_info(src_filepath)
                bbox = BoundingBox(tile_info['ulx'], tile_info['uly'], tile_info['lrx'], tile_info['lry'], epsg)
                create_template_raster(bbox, epsg, pixel_width, tile_composite_filepath, input_value=0, num_bands = n_bands)

def create_tilecomposites(raster_list, tile_infos, process_cpu_count = CPU_COUNT, mask_path = None):
    tile_extent_path_dict = {}
    for tile_info in tile_infos:
        if not Path(tile_info["tile_multiband_interpolated_composite"]).exists():
            tile_extent_path_dict[Path(tile_info['tile_folder']).name] = {'ulx': tile_info['ulx'],
                                                                 'uly': tile_info['uly'],
                                                                 'lrx': tile_info['lrx'],
                                                                 'lry': tile_info['lry'],
                                                                 'tile_folder': str(tile_info['tile_folder'])}

    if not len(tile_extent_path_dict.items()) == 0:
        tile_extent_path_dict_json = json.dumps(tile_extent_path_dict)

        subtasks = []
        for raster_list_item in raster_list:
            args = [str(raster_list_item), tile_extent_path_dict_json]
            subtasks.append(args)
        if mask_path is not None:
            arg = [str(mask_path), tile_extent_path_dict_json]
            subtasks.append(arg)

        p = Pool(process_cpu_count)
        p.map(cut_composite_totile_function, tuple(subtasks))
        p.close()  # Close the pool
        p.join()

def  create_interpolated_multiband(tile_infos, process_cpu_count = CPU_COUNT):
    subtasks = []
    for tile_info in tile_infos:
        if Path(tile_info["tile_multiband_interpolated_composite"]).exists(): continue
        args = [tile_info]
        log.debug(f"Adding task - {tile_info['row']} ,  {tile_info['col']}")
        subtasks.append(args)

    if len(subtasks) > 0:
        log.debug(f"Doing interpolating for {len(subtasks)}")
        p = Pool(process_cpu_count)
        p.map(generate_interpolated_multiband_tile, tuple(subtasks))
        p.close()  # Close the pool
        p.join()

def create_tiled_multiband(raster_list, work_dir, process_cpu_count = CPU_COUNT, mask_path = None):

    (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(raster_list[0])
    tile_infos = setup_tiles(xmin, ymax, xsize, ysize, pixel_width, n_bands,MAX_TILE_SIZE, work_dir, mask_path= mask_path)

    create_tilecomposites(raster_list, tile_infos, process_cpu_count=process_cpu_count, mask_path=mask_path)
    create_multiband_after_tiling(raster_list, tile_infos)
    return tile_infos

def interpolate_rasters(raster_list, work_dir, process_cpu_count = CPU_COUNT, do_recursive_mosaic = False, keep_multiband = True,
                        interpolated_filename = None, mask_path = None):

    tile_infos = create_tiled_multiband(raster_list, work_dir, process_cpu_count, mask_path = mask_path)

    create_interpolated_multiband(tile_infos, process_cpu_count=process_cpu_count)

    if interpolated_filename is not None:
        mosaic_filepath = work_dir.joinpath(interpolated_filename)
    else:
        mosaic_filepath = work_dir.joinpath('interpolate.tif')

    if keep_multiband:
        log.debug(f"Keeping multiband")
        if not mosaic_filepath.exists():
            if not do_recursive_mosaic:
                log.debug("Doing non recursive mosaic")
                mosaic_tile_tifs(tile_infos, mosaic_filepath, work_dir)
            else:
                do_recursive_tile_mosaic(tile_infos, mosaic_filepath, work_dir)
        log.debug(f"--X-- Done creating interpolated multiband tif --X--")
        return [mosaic_filepath]
    else:
        interpolated_list = []

        for raster_index, raster_list_item in enumerate(raster_list):
            merged_result_filepath = work_dir.joinpath(Path(raster_list_item).name)
            merged_vrt_filepath = work_dir.joinpath(Path(raster_list_item).name.replace(".tif", ".vrt"))

            merged_mask_result_filepath = work_dir.joinpath(f"mask_{Path(raster_list_item).name}")
            merged_mask_vrt_filepath = work_dir.joinpath(f"mask_{Path(raster_list_item).name.replace('.tif', '.vrt')}")


            if not merged_result_filepath.exists():
                log.debug("merged vrt filepath:{}".format(merged_vrt_filepath))
                args = ["gdalbuildvrt", "-b", f"{raster_index + 1}", f"{str(merged_vrt_filepath)}"]
                for tile_info in tile_infos:
                    tile_multiband_interpolated_path = tile_info["tile_multiband_interpolated_composite"]
                    args.append(str(tile_multiband_interpolated_path))

                log.debug("agrs:{}".format(args))
                subprocess.check_output(args)

                # Then convert the vrt to a merged GeoTiff file with compression.
                args = ["gdal_translate",
                            "-of", "GTiff",
                            "-co", "TILED=YES",
                            "-co", "COMPRESS=DEFLATE",
                            "-co", "BIGTIFF=YES",
                            str(merged_vrt_filepath),
                            str(merged_result_filepath)]
                subprocess.check_output(args)

            if not merged_mask_result_filepath.exists():
                log.debug("merged mask vrt filepath:{}".format(merged_mask_vrt_filepath))
                args = ["gdalbuildvrt", "-b", f"{raster_index + 1}", f"{str(merged_mask_vrt_filepath)}"]
                for tile_info in tile_infos:
                    tile_multiband_mask = tile_info["tile_multiband_mask"]
                    args.append(str(tile_multiband_mask))

                log.debug("agrs:{}".format(args))
                subprocess.check_output(args)

                # Then convert the vrt to a merged GeoTiff file with compression.
                args = ["gdal_translate",
                            "-of", "GTiff",
                            "-co", "TILED=YES",
                            "-co", "COMPRESS=DEFLATE",
                            "-co", "BIGTIFF=YES",
                            str(merged_mask_vrt_filepath),
                            str(merged_mask_result_filepath)]
                subprocess.check_output(args)

            interpolated_list.append((merged_result_filepath, merged_mask_result_filepath))
        return interpolated_list


def find_min_max(raster_filepath, master_mask=None, tile_size = 2000):

    if master_mask is not None:
        raster1 = gdal.Open(str(raster_filepath), gdal.GA_ReadOnly)
        raster2 = gdal.Open(str(master_mask), gdal.GA_ReadOnly)

        # Check if raster dimensions match
        if raster1.RasterXSize != raster2.RasterXSize or raster1.RasterYSize != raster2.RasterYSize:
            raise ValueError("Rasters have different dimensions.")

    mini = 0
    maxi = 0
    raster_ds = gdal.Open(str(raster_filepath), gdal.GA_ReadOnly)
    # Get the raster band
    raster_band = raster_ds.GetRasterBand(1)
    # Get the NoData value
    nodata_value = raster_band.GetNoDataValue()

    for x in range(0, raster_ds.RasterXSize, tile_size):
        for y in range(0, raster_ds.RasterYSize, tile_size):
            width = min(tile_size, raster_ds.RasterXSize - x)
            height = min(tile_size, raster_ds.RasterYSize - y)

            arr_r1 = raster_ds.GetRasterBand(1).ReadAsArray(x, y, width, height)
            if nodata_value is not None:
                arr_r1 = np.where(arr_r1 == nodata_value, np.nan, arr_r1)

            if master_mask is not None:
                master_chunk = raster2.GetRasterBand(1).ReadAsArray(x, y, width, height)
                arr_r1 = arr_r1[master_chunk==1]

            if numpy.nanmin(arr_r1) < mini:
                mini = numpy.nanmin(arr_r1)
            if numpy.nanmax(arr_r1) > maxi:
                maxi = numpy.nanmax(arr_r1)
    return mini, maxi







import os
from pathlib import Path
from typing import Optional
from osgeo import ogr, osr
from tondortools.tool import read_raster_info
from osgeo import gdal
import math
import subprocess
import logging
import shutil
import time
from copy import copy
import numpy as np
from multiprocessing import Pool
import geopandas as gpd
from shapely.ops import unary_union
import pandas as pd
from osgeo.gdalconst import GDT_Int16, GDT_Float32, GDT_Byte, GDT_Int16

try:
    from tondor.util.tiling import do_split_save_gpkg
    from tondor.util.tool import raster2array, create_maskedraster, read_raster_info, calculate_time
    from tondor.util.logging_config import init_logging, get_console_handler
except:
    from src.tondor.util.tiling import do_split_save_gpkg
    from src.tondor.util.tool import raster2array, create_maskedraster, read_raster_info, calculate_time
    from src.tondor.util.logging_config import init_logging, get_console_handler


import logging


init_logging()

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(get_console_handler())  # Add the common handler
log.propagate = False



USE_PARALLEL = True
CPU_COUNT = 7

def do_gpkg_reprojection(source_gpkg, target_gpkg, template_epsg):

    epsg =get_gpkg_epsg(template_epsg)

    if not Path(target_gpkg).exists():
        reproject_gpkg = ["ogr2ogr", "-f", "gpkg", "-t_srs", f"epsg:{epsg}", str(target_gpkg),
                          str(source_gpkg)]
        cmd_output = subprocess.run(reproject_gpkg, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        print("exit code {} --> {}".format(cmd_output.returncode, reproject_gpkg))

    return target_gpkg



def get_gpkg_epsg(gpkg_path):
    source = ogr.Open(str(gpkg_path), update=False)
    layer = source.GetLayer()
    epsg = layer.GetSpatialRef().GetAuthorityCode(None)
    source = None
    layer = None
    return epsg

def get_layer_name(gpkg_path):
    source = ogr.Open(str(gpkg_path), update=False)
    layer = source.GetLayer()
    layer_name = layer.GetName()
    return layer_name

def create_raster_bounds(raster, sourceSR, output_tiff):
    (xmin, ymax, RasterXSize, RasterYSize, pixel_width, _, epsg, _, _, raster_bbox) = read_raster_info(raster)

    # Define the input GeoPackage and output GeoTIFF
    mem_driver = gdal.GetDriverByName('Gtiff')
    mem_raster = mem_driver.Create(str(output_tiff), RasterXSize, RasterYSize,
                                   1, gdal.GDT_Byte)
    # Set the geotransform
    mem_raster.SetGeoTransform((xmin, pixel_width, 0, ymax, 0, -pixel_width))
    mem_raster.SetProjection(sourceSR.ExportToWkt())
    mem_band = mem_raster.GetRasterBand(1)
    mem_band.Fill(0)
    mem_raster = None

def get_parcel_positions(output_tiff, raster):
    ds_output = gdal.Open(str(output_tiff))
    outputarray = ds_output.GetRasterBand(1).ReadAsArray()
    outputarray = outputarray.astype(np.float32)
    arraymask = outputarray == 1
    positions = np.argwhere(arraymask == True)
    return positions

def create_parcelgpkg(sourceSR, feature, output_gpkg, raster=None):
    # Read parcel geometry
    feature_geom = feature.GetGeometryRef()
    if raster is not None:
        # Open data
        raster_ds = gdal.Open(str(raster))
        # Reproject vector geometry to same projection as raster
        targetSR = osr.SpatialReference()
        targetSR.ImportFromWkt(raster_ds.GetProjectionRef())
        coordTrans = osr.CoordinateTransformation(sourceSR, targetSR)
        feature_geom.Transform(coordTrans)

    # Create a memory layer from the feature
    mem_driver = ogr.GetDriverByName("GPKG")
    mem_ds = mem_driver.CreateDataSource(str(output_gpkg))

    if (feature_geom.GetGeometryName() == 'MULTIPOLYGON'):
        mem_layer = mem_ds.CreateLayer("parcel", sourceSR, ogr.wkbMultiPolygon)
    if (feature_geom.GetGeometryName() == 'POLYGON'):
        mem_layer = mem_ds.CreateLayer("parcel", sourceSR, ogr.wkbPolygon)
    new_field = ogr.FieldDefn('ID', ogr.OFTString)
    new_field.SetWidth(100)
    mem_layer.CreateField(new_field)

    mem_feature = ogr.Feature(mem_layer.GetLayerDefn())
    mem_feature.SetGeometry(feature.GetGeometryRef())
    mem_feature.SetField('ID', str(1))
    mem_layer.CreateFeature(mem_feature)

    feature_bbox = mem_feature.GetGeometryRef().GetEnvelope()
    mem_ds = None
    return feature_bbox

def get_parcel_locations(feature, sourceSR, raster, touch_status, work_dir):

    feature_fid= feature.GetFID()
    output_gpkg = Path(work_dir).joinpath(f"parcel_{feature_fid}.gpkg")
    output_tiff = Path(work_dir).joinpath(f"parcel_{feature_fid}.tif")

    if output_tiff.exists():
        output_tiff.unlink()

    if output_gpkg.exists():
        output_gpkg.unlink()

    feature_bbox = create_parcelgpkg(sourceSR, feature, output_gpkg, raster)

    create_raster_bounds(raster, sourceSR, output_tiff)
    create_maskedraster(output_gpkg, output_tiff, touch_status)
    parcel_position = get_parcel_positions(output_tiff, raster)

    if output_tiff.exists():
        output_tiff.unlink()

    if output_gpkg.exists():
        output_gpkg.unlink()
    return parcel_position

def parcelids_indices(object_path, raster, touch_status, work_dir, parcel_column=None):
    source = ogr.Open(str(object_path), update=False)
    layer = source.GetLayer()
    sourceSR = layer.GetSpatialRef()

    parcel_location_dict = {}

    for parcel in layer:
        parcel_locations = get_parcel_locations(parcel, sourceSR, raster, touch_status, Path(work_dir))
        if parcel_column is not None:
            parcel_location_dict[str(parcel.GetField(parcel_column))] = parcel_locations
        else:
            parcel_location_dict[str(parcel.GetFID())] = parcel_locations

    return parcel_location_dict

def update_object_count(object_path, parcel_location_dict, column_name = "count", parcel_column=None):
    source = ogr.Open(str(object_path), update=True)
    layer = source.GetLayer()
    sourceSR = layer.GetSpatialRef()

    ## set pixel count
    new_field = ogr.FieldDefn(column_name, ogr.OFTReal)
    new_field.SetWidth(5)
    layer.CreateField(new_field)

    for parcel in layer:
        if parcel_column is not None:
            parcel_name = str(parcel.GetField(parcel_column))
        else:
            parcel_name = str(parcel.GetFID())

        parcel_locations = parcel_location_dict[parcel_name]
        parcel_count = parcel_locations.shape[0]
        parcel.SetField(column_name, parcel_count)
        layer.SetFeature(parcel)
    source = None
    layer = None
    log.debug(f" --- {object_path} done ---")

##########################
def create_work_dirs(gpkg_paths, work_dir):

    gpkg_workdir_tuple_list = []
    for gpkg_path_item in gpkg_paths:
        gpkg_path_item_name = Path(gpkg_path_item).stem
        gpkg_work_dir = work_dir.joinpath(gpkg_path_item_name)
        os.makedirs(gpkg_work_dir, exist_ok=True)
        gpkg_workdir_tuple_list.append((str(gpkg_path_item), str(gpkg_work_dir)))
    return gpkg_workdir_tuple_list


def count_pixel_subgpkg(input_data):
    gpkg_path = Path(input_data[0])
    template_raster = Path(input_data[1])
    touch_status = input_data[2]
    work_dir = input_data[3]
    parcel_column = input_data[4]
    column_name = input_data[5]
    if "_NODATA.txt" in gpkg_path.name:
        return

    parcel_location_dict = parcelids_indices(gpkg_path, template_raster, touch_status, work_dir, parcel_column)
    update_object_count(gpkg_path, parcel_location_dict, column_name, parcel_column=parcel_column)

##########################

class ObjectLayer:
    def __init__(self, config):
        self.config = config
        self.touch_status = None
        self.pixel_size = Optional[int]

        self.origin = None
        self.target_epsg = None

        self.parse_config(self.config)
        self.setup_origin(self.config.get("origin", None))

        self.object_layer = None
        self.parcel_column = None
        self.local_path = None
        self.local_path_status = False
        
        self.parcel_location_dict = None

    @property
    def parcel_ids(self):
        if self.parcel_location_dict is None:
            raise Exception("call set_parcelids_indices function before calling this function")
        return self.parcel_location_dict.keys()

    @property
    def object_path(self):
        if self.local_path is not None:
            return self.local_path

        if self.object_layer is not None:
            return self.object_layer

        raise Exception(f"object layer not set.")

    def set_object_layer(self, gpkg_filepath):
        if not Path(gpkg_filepath).exists():
            raise Exception(f"{gpkg_filepath} does not exists.")
        self.object_layer = gpkg_filepath

    def set_check_parcel_column(self, column_name):
        CLASS_ATTRIBUTE = column_name
        source = ogr.Open(str(self.object_layer), update=False)
        layer = source.GetLayer()
        layer_defn = layer.GetLayerDefn()
        field_names = [layer_defn.GetFieldDefn(i).GetName() for i in range(layer_defn.GetFieldCount())]
        if CLASS_ATTRIBUTE not in field_names:
            msg = "File {:s} does not have expected attribute {:s}".format(str(self.object_layer), CLASS_ATTRIBUTE)
            raise Exception(msg)
        else:
            self.parcel_column = column_name
            return True

    def set_rasternodata_value(self, nodata_value):
        self.nodata_value = nodata_value

    def copy_totmp(self, tmp_dir):
        self.local_path = Path(tmp_dir).joinpath(Path(self.object_layer).name)
        shutil.copy(self.object_layer, self.local_path)
        self.local_path_status = True

    def set_local_path(self):
        self.local_path = self.object_path
        self.local_path_status = True
        
    def reproject(self):
        if self.local_path_status == False: raise Exception(f"copy to local path first")
        source = ogr.Open(str(self.local_path), update=False)
        layer = source.GetLayer()
        input_gpkg_epsg = layer.GetSpatialRef().GetAuthorityCode(None)
        if int(input_gpkg_epsg) != int(self.target_epsg):
            gpkg_copy = Path(self.local_path).parent.joinpath(str(self.local_path.name).replace('.gpkg', '_copy.gpkg'))
            shutil.copy(self.local_path, gpkg_copy)
            self.local_path.unlink()
            reproject_gpkg = ["ogr2ogr", "-f", "gpkg", "-t_srs", f"epsg:{self.target_epsg}", str(self.local_path), str(gpkg_copy)]
            cmd_output = subprocess.run(reproject_gpkg, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            log.info("exit code {} --> {}".format(cmd_output.returncode, reproject_gpkg))
            gpkg_copy.unlink()

    def parse_config(self, config):
        all_in_status = config.get("touch_status", "all_on")
        if not all_in_status in ["all_in", "all_touch", "all_on"]:
            raise Exception(f"touch status can be only all_in, all_touch, all_on.")

        self.touch_status = all_in_status

        pixel_size = config.get("pixel_size", 10)
        self.pixel_size = int(pixel_size)

    def setup_origin(self, origin=None, raster=None):
        if origin is not None:
            self.origin = origin["coordinates"]
            self.target_epsg = origin["epsg"]

        if raster is not None:
            (xmin, ymax, _, _, _, _, epsg, _, _, _) = read_raster_info(
                raster)
            self.origin = (xmin, ymax)
            self.target_epsg = epsg

    def check_origin(self, raster=None):
        if raster == None and self.origin == None:
            raise Exception(f"Origin is not set. Extraction will not be correct")

        if raster is not None:
            (xmin, ymax, _, _, _, _, epsg, _, _, _) = read_raster_info(
                raster)
            self.origin = (xmin, ymax)
            self.target_epsg = epsg

    def setup_raster_extent(self, feature_bbox, output_gpkg):
        pass

    ########################################
    ########################################
    ########################################
    def find_nearest_vertex(self, point, pixel_size, position):
        # define origin and pixel size
        origin = self.origin  # example origin, replace with your own values

        if position == 'upperleft':

            # calculate the row and column of the pixel
            row = math.ceil((point[1] - origin[1]) / pixel_size) + 1
            col = math.ceil((point[0] - origin[0]) / pixel_size) - 1

            # calculate the nearest top left vertex of the pixel
            vertex_x = origin[0] + col * pixel_size
            vertex_y = origin[1] + row * pixel_size

        elif position == 'bottomright':
            row = math.floor((point[1] - origin[1]) / pixel_size) - 1
            col = math.floor((point[0] - origin[0]) / pixel_size) + 1

            # calculate the nearest bottom right vertex of the pixel
            vertex_x = origin[0] + (col + 1) * pixel_size
            vertex_y = origin[1] + (row + 1) * pixel_size

        return (vertex_x, vertex_y)

    def create_raster_parcelbounds(self, pixel_size, feature_bbox, sourceSR, output_tiff):

        featurexmin = feature_bbox[0]
        featurexmax = feature_bbox[1]
        featureymin = feature_bbox[2]
        featureymax = feature_bbox[3]

        point = (featurexmin, featureymax)
        (xmin, ymax) = self.find_nearest_vertex(point, pixel_size, position='upperleft')
        point = (featurexmax, featureymin)
        (xmax, ymin) = self.find_nearest_vertex(point, pixel_size, position='bottomright')

        xOrigin = self.origin[0]
        yOrigin = self.origin[1]
        # Specify offset and rows and columns to read
        xoff = int((xmin - xOrigin) / pixel_size)
        yoff = int((yOrigin - ymax) / pixel_size)
        xcount = int((xmax - xmin) / pixel_size) + 1
        ycount = int((ymax - ymin) / pixel_size) + 1

        # Define the input GeoPackage and output GeoTIFF
        mem_driver = gdal.GetDriverByName('Gtiff')
        mem_raster = mem_driver.Create(str(output_tiff), xcount, ycount,
                                       1, gdal.GDT_Byte)
        # Set the geotransform
        mem_raster.SetGeoTransform((xmin, pixel_size, 0, ymax, 0, -pixel_size))
        mem_raster.SetProjection(sourceSR.ExportToWkt())
        mem_band = mem_raster.GetRasterBand(1)
        mem_band.Fill(0)
        mem_raster = None
        offset = (xoff, yoff)
        return output_tiff, offset

    def get_raster_values(self, output_tiff, raster, offset):
        ds_output = gdal.Open(str(output_tiff))
        outputarray = ds_output.GetRasterBand(1).ReadAsArray()
        outputarray = outputarray.astype(np.float32)
        arraymask = outputarray == 1

        # Read raster as arrays
        (_, _, RasterXSize, RasterYSize, _, _, _, _, _, _) = read_raster_info(output_tiff)
        raster_ds = gdal.Open(str(raster))
        banddataraster = raster_ds.GetRasterBand(1)
        dataraster = banddataraster.ReadAsArray(offset[0], offset[1], RasterXSize, RasterYSize).astype(np.float32)
        return dataraster[arraymask]

    def aggregate_raster_parcel(self, feature, pixel_size, sourceSR, config, raster=None, work_dir=None):

        feature_fid= feature.GetFID()
        output_gpkg = Path(work_dir).joinpath(f"parcel_{feature_fid}.gpkg")
        output_tiff = Path(work_dir).joinpath(f"parcel_{feature_fid}.tif")

        if output_gpkg.exists():
            output_gpkg.unlink()
        if output_tiff.exists():
            output_tiff.unlink()

        self.check_origin(raster)
        feature_bbox = create_parcelgpkg(sourceSR, feature, output_gpkg, raster)

        output_tiff, offset = self.create_raster_parcelbounds(pixel_size, feature_bbox, sourceSR, output_tiff)

        create_maskedraster(output_gpkg, output_tiff, self.touch_status)
        raster_values = self.get_raster_values(output_tiff, raster, offset)

        if output_tiff.exists():
            output_tiff.unlink()
        if output_gpkg.exists():
            output_gpkg.unlink()

        return raster_values

    ########################################
    ### functionalities ###########
    ########################################

    def set_parcelids_indices(self, work_dir, raster):
        parcel_location_dict = parcelids_indices(self.object_path, raster, self.touch_status, work_dir, self.parcel_column)
        self.parcel_location_dict = parcel_location_dict

    def set_pixel_count(self):
        if self.parcel_location_dict is None:
            raise Exception("call set_parcelids_indices function before calling this function")
        update_object_count(self.object_path, self.parcel_location_dict, self.parcel_column)

    def get_parcelids(self):
        if self.parcel_location_dict is None:
            raise Exception("call set_parcelids_indices function before calling this function")
        return self.parcel_location_dict.keys()

    def get_raster_parcel_values(self, parcel_id, input_array):
        mask_index = self.get_parcel_indices(parcel_id)

        if isinstance(input_array, np.ndarray):
            parcel_values = input_array[mask_index[:, 0], mask_index[:, 1]]
            return parcel_values

        if isinstance(input_array, Path):
            raster = gdal.Open(str(input_array))
            nodata_value = raster.GetRasterBand(1).GetNoDataValue()

            raster_array = raster2array(str(input_array))
            raster_array[raster_array == nodata_value] = np.nan
            parcel_values = raster_array[mask_index[:, 0], mask_index[:, 1]]
            return parcel_values

        raise Exception(f"neither array nor path")

    def get_parcel_indices(self, parcel_id):
        return self.parcel_location_dict[parcel_id]

    ########################################
    ##### parallel pixel count ########
    ########################################

    def count_pixels_parallel(self, work_dir, raster_template, column_name= "count"):
        layer_name = get_layer_name(self.local_path)
        gpkg_paths = do_split_save_gpkg(self.local_path, self.local_path.name, work_dir, subtiles_count=14)
        gpkg_workdir = create_work_dirs(gpkg_paths, work_dir)

        subtasks = []
        for gpkg_workdir_item in gpkg_workdir:
            args = [gpkg_workdir_item[0], raster_template, self.touch_status, gpkg_workdir_item[1], self.parcel_column, column_name]
            subtasks.append(args)

        p = Pool(CPU_COUNT)
        p.map(count_pixel_subgpkg, tuple(subtasks), chunksize=1)


        self.local_path.unlink()
        for gpkg_workdir_item in gpkg_workdir:
            tile_gpkg_path = gpkg_workdir_item[0]
            org_cmd = ["ogr2ogr", "-f", "gpkg", "-append", "-nln", layer_name,
                       str(self.local_path), str(tile_gpkg_path)]
            cmd_output = subprocess.run(org_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            log.info("exit code {} --> {}".format(cmd_output.returncode, org_cmd))
        log.info(f"{self.local_path} created after pixel count")

    def process_gpkg(self, column_name = "L1_code"):
        source_gpkg = self.local_path
        result_gpkg = self.local_path.parent.joinpath(f"{self.local_path.stem}_merged.gpkg")
        process_gpkg(source_gpkg, result_gpkg, column_name=column_name)

    def remove_overlap(self):
        if self.local_path_status == False: raise Exception(f"copy to local path first")
        shutil.copy(self.local_path, self.local_path.parent.joinpath(self.local_path.name.replace(".gpkg", "_ori.gpkg")))
        gdf = gpd.read_file(self.local_path)

        invalid_mask = ~gdf.geometry.is_valid
        gdf.loc[invalid_mask, 'geometry'] = gdf.loc[invalid_mask, 'geometry'].buffer(0)

        # Iterate through each polygon and subtract it from the rest to remove overlaps
        non_overlapping_gdf = gpd.GeoDataFrame(columns=gdf.columns)
    
        for index, row in gdf.iterrows():
    
            neighbors = gdf.drop(index)[gdf.touches(row['geometry']) | gdf.intersects(row['geometry'].buffer(1))]
            if len(neighbors) ==0:
                non_overlapping_gdf = gpd.GeoDataFrame(pd.concat([non_overlapping_gdf, gpd.GeoDataFrame([row])], ignore_index=True))
                continue
            polygon = row.geometry
            other_polygons = neighbors.geometry.unary_union
            non_overlapping_polygon = polygon.difference(other_polygons)
            if not non_overlapping_polygon.is_empty:
                new_row = row.copy()
                new_row.geometry = non_overlapping_polygon
                non_overlapping_gdf = gpd.GeoDataFrame(pd.concat([non_overlapping_gdf, gpd.GeoDataFrame([new_row])], ignore_index=True))
            gdf.drop(index, inplace=True)
    
        non_overlapping_gdf.set_crs(gdf.crs, inplace=True)
        self.local_path.unlink()
        non_overlapping_gdf.to_file(self.local_path, driver="GPKG")


def recode_gpkg_column(gpkg_path, conversion_table_path,
                       conversion_table_original_column, conversion_table_target_column,
                       gpkg_column):

    # Read the GeoPackage file
    gdf = gpd.read_file(gpkg_path)

    # Read the conversion table CSV
    conversion_df = pd.read_csv(conversion_table_path, encoding_errors='ignore')

    gdf[f"{gpkg_column}_str"] = gdf[f"{gpkg_column}"].astype(str)
    conversion_dict = dict(
        zip(conversion_df[conversion_table_original_column].astype(str), conversion_df[conversion_table_target_column]))
    # Apply the conversion to the specified column in the GeoDataFrame
    gdf[conversion_table_target_column] = gdf[f"{gpkg_column}_str"].map(conversion_dict)
    gdf.drop(columns=[f"{gpkg_column}_str"], inplace=True)

    gpkg_path.unlink()
    # Save the modified GeoDataFrame to a new GeoPackage file
    gdf.to_file(gpkg_path, driver="GPKG")

    log.info(f"Recode completed: {gpkg_path}. {gpkg_column} --> {conversion_table_target_column}")


def group_polygons_assigngroupid(gdf_input):
    gdf = copy(gdf_input)

    invalid_mask = ~gdf.geometry.is_valid
    gdf.loc[invalid_mask, 'geometry'] = gdf.loc[invalid_mask, 'geometry'].buffer(0)

    # Create a new column to store the group ID for each polygon
    gdf['group_id'] = -1

    group_id = 0
    # Iterate over each row in the GeoDataFrame
    for idx, row in gdf.iterrows():

        # Get the group ID of the current row
        if row['group_id'] != -1:
            row_group_id = row['group_id']
        else:
            row_group_id = group_id
            group_id += 1

        # Find all polygons that touch or intersect the current row's geometry
        neighbors = gdf[gdf.touches(row['geometry']) | gdf.intersects(row['geometry'].buffer(1))]

        # Check if any neighbors have a different group ID
        neighbor_group_ids = neighbors.loc[neighbors['group_id'] != -1, 'group_id']
        if not neighbor_group_ids.empty:
            # Get the unique group IDs of the neighbors
            unique_group_ids = neighbor_group_ids.unique()

            # Assign the current row's group ID to the neighbors with different group IDs
            gdf.loc[neighbors.index, 'group_id'] = row_group_id

            # Update the group ID for all polygons with the unique group IDs
            gdf.loc[gdf['group_id'].isin(unique_group_ids), 'group_id'] = row_group_id

        else:
            # Assign the current row's group ID to the neighbors with the default group ID
            gdf.loc[neighbors.index, 'group_id'] = row_group_id

    grouped_geoms = []
    for unique_group_id in gdf.group_id.unique():
        gdf_group_id = gdf.loc[gdf.group_id == unique_group_id]
        touching_geoms = []
        for idx, row in gdf_group_id.iterrows():
            touching_geoms.append(row)
        grouped_geoms.append(touching_geoms)
    return grouped_geoms

def convert_list_gdf(merged_geoms_list):
    df= pd.DataFrame(merged_geoms_list)
    if len(df) > 0:
        gdf = gpd.GeoDataFrame(df, geometry='geometry')
    else:
        gdf = gpd.GeoDataFrame()
    return gdf

def merge_geoms(grouped_geoms):
    merged_attribute_list = []
    for group_geom in grouped_geoms:
        if len(group_geom) > 1:
            merged_geom = unary_union([geom_item.geometry for geom_item in group_geom])
            tol = 3
            corrected_geom = merged_geom.buffer(tol).buffer(-tol)

            merged_attribute = copy(group_geom[0])
            merged_attribute.geometry = corrected_geom
            merged_attribute_list.append(merged_attribute)
        else:
            merged_attribute = copy(group_geom[0])
            merged_attribute_list.append(merged_attribute)

    return  merged_attribute_list

def perform_parcel_merge_subgpkg(input_data):
    gpkg_path = Path(input_data[0])
    gpkg_merged_path = Path(input_data[1])

    gdf = gpd.read_file(gpkg_path)

    group_withid = group_polygons_assigngroupid(gdf)
    merged_geoms_list_group_id = merge_geoms(group_withid)
    gdf_group_id = convert_list_gdf(merged_geoms_list_group_id)
    gdf_group_id.to_file(gpkg_merged_path, driver="GPKG")

def process_gpkg(source_gpkg, result_gpkg, column_name = "SAMAS"):
    master_df = gpd.read_file(source_gpkg)
    group_df = gpd.GeoDataFrame(columns=master_df.columns)
    group_df_groupid = gpd.GeoDataFrame(columns=master_df.columns)

    subtasks = []
    files_to_merge = []
    for value in master_df[column_name].unique():
        group = master_df.loc[master_df[column_name] == value]
        group_gpkg_path = source_gpkg.parent.joinpath(f"{source_gpkg.stem}_{str(value)}{source_gpkg.suffix}")
        group_gpkg_merged_path = source_gpkg.parent.joinpath(f"{source_gpkg.stem}_{str(value)}_merged{source_gpkg.suffix}")
        group.to_file(str(group_gpkg_path), driver="GPKG")
        #perform_parcel_merge_subgpkg([group_gpkg_path, group_gpkg_merged_path])

        args = [str(group_gpkg_path), str(group_gpkg_merged_path)]
        subtasks.append(args)

    p = Pool(CPU_COUNT)
    p.map(perform_parcel_merge_subgpkg, tuple(subtasks), chunksize=1)


    for gpkg_workdir_item in gpkg_workdir:
        args = [gpkg_workdir_item[0], raster_template, self.touch_status, gpkg_workdir_item[1], self.parcel_column,
                column_name]
        subtasks.append(args)
    
    ori_gdf = gpd.read_file(source_gpkg)  # specify the layer name if necessary

    gdf = ori_gdf.dissolve(by=column_name)
    
    # Find unique attributes and initialize an empty GeoDataFrame for results
    unique_attrs = gdf[column_name].unique()
    result_gdf = gpd.GeoDataFrame(columns=gdf.columns, crs=gdf.crs)

    for attr in unique_attrs:
        # Filter polygons with the current attribute
        current_polygons = gdf[gdf[column_name] == attr]

        # Dissolve into a single polygon if possible
        dissolved = current_polygons.dissolve()

        # Subtract overlaps with different attributes
        other_polygons = gdf[gdf[attribute] != attr]
        for _, other_row in other_polygons.iterrows():
            dissolved['geometry'] = dissolved['geometry'].apply(lambda x: x.difference(other_row['geometry']))

        result_gdf = result_gdf.append(dissolved.reset_index(), ignore_index=True)
    
    result_gdf.to_file(str(result_gpkg), driver='GPKG')
    return result_gdf


def inner_buffer(geometry, buffer_distance):
    return geometry.buffer(-buffer_distance)

def create_gdf_inner_buffer(input_gdf, buffer_factor):
    pixel_size = 1
    buffer_distance = buffer_factor*pixel_size
    input_gdf["geometry"] = input_gdf["geometry"].apply(lambda geom: inner_buffer(geom, buffer_distance))
    return input_gdf

def create_gpkg_inner_buffer(input_gpkg, output_gpkg, buffer_factor):
    gdf = gpd.read_file(input_gpkg)

    output_gdf = create_gdf_inner_buffer(gdf, buffer_factor)
    
    output_gdf.to_file(str(output_gpkg), driver="GPKG")
    log.debug(f"Inner buffer applied and saved {output_gpkg}")
    
    
def count_vector_pixel(gpkg_path, template_raster, work_dir, touch_status ="all_in", field_name = 'pixel_count', data_type = GDT_Int16):
    (xmin, ymax, RasterXSize, RasterYSize, pixel_width, _, epsg, _, _, _) = read_raster_info(
        template_raster)

    raster_parcel_config = {"touch_status": touch_status, "pixel_size": pixel_width}
    ObjectLayer_instance = ObjectLayer(raster_parcel_config)
    ObjectLayer_instance.set_object_layer(gpkg_path)
    ObjectLayer_instance.set_local_path()
    ObjectLayer_instance.setup_origin(raster=template_raster)
    ObjectLayer_instance.reproject()

    source = ogr.Open(str(gpkg_path), update=True)
    layer = source.GetLayer()
    sourceSR = layer.GetSpatialRef()

    new_field = ogr.FieldDefn(field_name, ogr.OFTReal)
    new_field.SetWidth(5)
    layer.CreateField(new_field)

    raster_ds = gdal.Open(str(template_raster))
    banddataraster = raster_ds.GetRasterBand(1)
    banddataarray = banddataraster.ReadAsArray()
    banddataarray = banddataarray.astype(np.float32)
    raster_allinpixel_values = np.zeros_like(banddataarray)

    log.debug(f"-- Getting pixels for {layer.GetFeatureCount()} parcels--")
    start_time = time.time()
    ObjectLayer_instance.set_parcelids_indices(work_dir, template_raster)
    calculate_time(start_time)

    for parcel in layer:
        if ObjectLayer_instance.parcel_column is None:
            parcel_name = str(parcel.GetFID())
        else:
            parcel_name = str(parcel.GetField(ObjectLayer_instance.parcel_column))

        parcel_locations = ObjectLayer_instance.get_parcel_indices(parcel_name)
        parcel_count = parcel_locations.shape[0]
        parcel.SetField(field_name, parcel_count)
        layer.SetFeature(parcel)

    source = None    
    layer = None
    raster_ds = None


import numpy
import shutil
from pathlib import Path
import subprocess
import logging
import numpy as np
import os

from tondortools.tool import read_raster_info, save_raster, mosaic_tifs, run_subprocess
import osgeo.gdalconst as gdalconst
try:
    from src.tondor.util.tool import setup_tiles, save_raster_template, find_file_in_archives, clip_raster_to_extent, reproject_multibandraster_toextent, read_raster_info, raster2array
    from src.tondor.util.stat import merge_tiles
    from src.tondor.util.geo import BoundingBox
    from src.tondor.util.logging_config import init_logging, get_console_handler
except:
    from tondor.util.tool import setup_tiles, save_raster_template, find_file_in_archives, clip_raster_to_extent, reproject_multibandraster_toextent, read_raster_info, raster2array
    from tondor.util.stat import merge_tiles
    from tondor.util.geo import BoundingBox
    from tondor.util.logging_config import init_logging, get_console_handler
    
from osgeo import gdal
from osgeo.gdalconst import GA_ReadOnly, GDT_Float32, GDT_Byte, GDT_Int16
from scipy.ndimage import binary_erosion

MAX_TILE_SIZE = 4000

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(get_console_handler()) 
log.propagate = False

def reclassify_subtiles(master_lc_mask, tile_infos, epsg, nodata_value):
    dataset = gdal.Open(str(master_lc_mask), GA_ReadOnly)
    for tile_info in tile_infos:
        start_x = tile_info["x_offset"]
        start_y = tile_info["y_offset"]
        read_width = tile_info["width"]
        read_height = tile_info["height"]
        ulx = tile_info["ulx"]
        uly = tile_info["uly"]

        # Read the chunk from the raster
        chunk = dataset.ReadAsArray(start_x, start_y, read_width, read_height)
        chunk_mask = numpy.ones_like(chunk)
        chunk_mask[chunk==nodata_value] = 0

        save_raster(chunk_mask, tile_info["tile_multiband_interpolated_composite"], "GTiff", int(epsg),
                    ulx, uly,
                    tile_info["pixel_size"], data_type=GDT_Int16)




def reclassify_raster(master_lc_mask, relcassified_path, work_dir):

    raster_ds = gdal.Open(str(master_lc_mask))
    raster_nodata = raster_ds.GetRasterBand(1).GetNoDataValue()

    (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(master_lc_mask)
    tile_infos = setup_tiles(xmin, ymax, xsize, ysize, pixel_width, n_bands,MAX_TILE_SIZE, work_dir)

    reclassify_subtiles(master_lc_mask, tile_infos, epsg, raster_nodata)
    raster_list = []
    for tile_info in tile_infos:
        raster_list.append(tile_info["tile_multiband_interpolated_composite"])

    print(f"Mosaic patches for {relcassified_path}")
    mosaic_tifs(raster_list, Path(relcassified_path))

def create_sieved_raster(input_raster, output_raster, sieve=8, diagonal_flag = True):
    (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(input_raster)

    if diagonal_flag: diag_num = "-8"
    else: diag_num = "-4"
    sieve_args = ["gdal_sieve.py", #"-mask", str(mask_tif),
                  diag_num, "-nomask",
                  "-st", str(sieve),
                  str(input_raster), str(output_raster)]
    cmd_output = subprocess.run(sieve_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    log.debug("exit code {} --> {}".format(cmd_output.returncode, sieve_args))

def perform_binary_erosion(input_raster, output_raster,dilate_iterations=3, invert=False):
    (xmin, ymax, xsize, ysize, pixel_width, projection, epsg, data_type, n_bands, _) = read_raster_info(input_raster)

    ds_raster = gdal.Open(str(input_raster))
    raster_array = ds_raster.GetRasterBand(1).ReadAsArray()
    if invert:
        raster_array = numpy.logical_not(raster_array).astype(int)

    raster_nodata = ds_raster.GetRasterBand(1).GetNoDataValue()
    dilated_array = binary_erosion(raster_array, iterations=dilate_iterations, border_value=1).astype(
        raster_array.dtype)
    if invert:
        dilated_array = numpy.logical_not(dilated_array).astype(int)

    save_raster_template(input_raster, output_raster, dilated_array, GDT_Byte, raster_nodata)



def do_bioregion_mosaic(mosaic_path, bioregion_tiles, year, archive_roots, work_dir, pca_dimension=12):

    if not mosaic_path.exists():
        mosaic_input_list = []
        for bioregion_tilename, site_selection in bioregion_tiles.items():

            tile_stacked_filename = f"OPT_{bioregion_tilename}_{year}_{pca_dimension}.tif"
            opt_composite_site_relpath = Path(str(year)).joinpath(bioregion_tilename, 'interpolated_optcomposite', tile_stacked_filename)
            interpolated_raster = find_file_in_archives(opt_composite_site_relpath, archive_roots, error_on_missing=False)

            work_dir_tile = work_dir.joinpath(tile_stacked_filename)
            if not work_dir_tile.exists():
                if interpolated_raster is None:

                    # first look for dim 12 month
                    tile_fullstacked_filename = f"OPT_{bioregion_tilename}_{year}_12.tif"
                    opt_composite_site_relpath = Path(str(year)).joinpath(bioregion_tilename, 'interpolated_optcomposite',
                                                                          tile_fullstacked_filename)
                    interpolated_fulldim_raster = find_file_in_archives(opt_composite_site_relpath, archive_roots,
                                                                error_on_missing=False)
                    # if dim 12 month then do interpolation of stacked 4 month
                    if interpolated_fulldim_raster is None:
                        interpolated_raster_input_list = []
                        opt_raster_list = locate_tile_optcomposites(bioregion_tilename, year, archive_roots)
                        for opt_raster_list_index, opt_raster_list_item in enumerate(opt_raster_list):
                            if site_selection['selection'][opt_raster_list_index] == '1':
                                interpolated_raster_input_list.append(opt_raster_list_item)
                        log.debug(f"creating interpolated stack from {interpolated_raster_input_list}")
                        interpolate_rasters(interpolated_raster_input_list, work_dir_tile.parent,
                                                                                5, False, True, interpolated_filename=tile_stacked_filename)
                    else:
                        band_selection_list = []
                        site_selection_string = site_selection['selection']
                        interpolated_fulldim_raster_ds = gdal.Open(str(interpolated_fulldim_raster))
                        interpolated_fulldim_raster_bandcount = interpolated_fulldim_raster_ds.RasterCount
                        n_band = interpolated_fulldim_raster_bandcount/len(site_selection_string)
                        for site_selection_string_index, site_selection_string_item in enumerate(site_selection_string):
                            band_1stindex_array = np.arange(6)
                            if site_selection_string_item == '1':
                                band_nthindex_array = band_1stindex_array + (site_selection_string_index*n_band) + 1
                                band_selection_list.extend(list(band_nthindex_array))
                                log.debug(f"adding stack list {band_nthindex_array}")
                        log.debug(f"Calculated stacked raster with {band_selection_list}")

                        tile_vrt_filepath = work_dir.joinpath(tile_stacked_filename.replace('.tif', '.vrt'))

                        vrt_stack = []
                        for band_selection_list_item in band_selection_list:
                            tile_vrt_band_filepath = work_dir.joinpath(tile_stacked_filename.replace('.tif', f'band{band_selection_list_item}.vrt'))
                            gdalbuildvrt_cmd = ["gdalbuildvrt",
                                                "-b", str(band_selection_list_item),
                                                str(tile_vrt_band_filepath), str(interpolated_fulldim_raster)]
                            cmd_output = subprocess.run(gdalbuildvrt_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                            vrt_stack.append(tile_vrt_band_filepath)

                        args = ["gdalbuildvrt", "-separate", str(tile_vrt_filepath)]
                        for vrt_stack_item in vrt_stack:
                            args.append(vrt_stack_item)
                        subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

                        cmd_output = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        log.debug(f"exit code {cmd_output.returncode} --> {args}")

                        args = ["gdal_translate",
                                "-of", "GTiff",
                                "-co", "TILED=YES",
                                "-co", "COMPRESS=DEFLATE",
                                "-co", "BIGTIFF=YES",
                                str(tile_vrt_filepath),
                                str(work_dir_tile)]
                        cmd_output = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        log.debug(f"exit code {cmd_output.returncode} --> {args}")

                else:
                    shutil.copy(interpolated_raster, work_dir_tile)
                    log.debug(f"{interpolated_raster} --> {work_dir_tile}")

            interpolated_raster_filepath_list = [work_dir_tile]
            mosaic_input_list.extend([str(interpolated_raster_item) for interpolated_raster_item in interpolated_raster_filepath_list])

        merge_tiles(mosaic_input_list, mosaic_path.name, mosaic_path.parent, method='average')
    return mosaic_path


def normalize_stacked_raster(stacked_raster, normalized_raster, number_of_bands):
    pass

def compress_lzw(input_file, output_file):
    args = [
        'gdal_translate',
        "-co", "TILED=YES",
        "-co", "COMPRESS=LZW",
        "-co", "BIGTIFF=YES",
        str(input_file),
        str(output_file)
    ]
    cmd_output = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    log.debug(f"exit code {cmd_output.returncode} --> {args}")


class Raster_tiling():
    def __init__(self):
        self.bbox = None
        self.rasterxsize = None
        
    def set_clip_extent(self, xmin, xmax, ymin, ymax, epsg):
        self.bbox = BoundingBox(xmin, ymin, xmax, ymax, epsg)
        return self.bbox
    
    def set_bbox(self, bbox):
        self.bbox = bbox
    
    def check_clipped_raster(self, dst_filepath):
        (xmin, ymax, RasterXSize, RasterYSize, pixel_width, projection, epsg, datatype, n_bands, imagery_extent_box) = read_raster_info(dst_filepath)
        if self.rasterxsize is None:
            self.rasterxsize = RasterXSize
            self.rasterysize = RasterYSize
            self.xmin = xmin
            self.ymax = ymax
            self.pixel_width = pixel_width
            self.epsg = epsg
        else:

            assert self.xmin == xmin
            assert self.ymax == ymax
            assert self.epsg == epsg
            
            pixel_factor = np.divide(self.pixel_width, pixel_width)
            assert self.rasterxsize*pixel_factor == RasterXSize
            assert self.rasterysize*pixel_factor == RasterYSize
            
    def clip_sen_tif_to_extent(self, src_filepath, dst_filepath, create_new = False):
        if not dst_filepath.exists() or create_new:
            clip_raster_to_extent(src_filepath, dst_filepath, self.bbox.xmin, self.bbox.xmax, self.bbox.ymin, self.bbox.ymax)
        self.check_clipped_raster(dst_filepath)

    def warp_sen_tif_to_extent(self, src_filepath, dst_filepath, pixel_width, method='near', create_new = False):
        if not dst_filepath.exists() or create_new:
            reproject_multibandraster_toextent(src_filepath, dst_filepath, self.bbox.epsg, pixel_width, self.bbox.xmin, self.bbox.xmax, self.bbox.ymin, self.bbox.ymax, method=method)
        self.check_clipped_raster(dst_filepath)

    def set_raster_params(self, raster_path):
        (xmin, ymax, RasterXSize, RasterYSize, pixel_width, projection, epsg, datatype, n_bands, imagery_extent_box) = read_raster_info(dst_filepath)
        if self.rasterxsize is None:
            self.rasterxsize = RasterXSize
            self.rasterysize = RasterYSize
            self.xmin = xmin
            self.ymax = ymax
            self.pixel_width = pixel_width
            self.epsg = epsg
            
    def rasterize_gpkg(self):
        pass
        
            
def create_individule_tifs_from_bands(tif_path, band_list, work_dir):
    band_tifpath_dict = {}
    for band in band_list:
        
        band_tif_filename = f'{Path(tif_path).stem}_band_{band}.tif'
        band_tif_filepath = Path(work_dir).joinpath(band_tif_filename)
        if not band_tif_filepath.exists():
            
            vrt_file_basename = f'{Path(tif_path).stem}_band_{band}.vrt'
            vrt_filepath = Path(work_dir).joinpath(vrt_file_basename)
            gdalbuildvrt_cmd = ["gdalbuildvrt",
                                "-b", str(band),
                                str(vrt_filepath), tif_path]
            run_subprocess(gdalbuildvrt_cmd, work_dir)
    
    
    
            translate_to_tif = ["gdal_translate",
                                "-of", "GTiff", str(vrt_filepath), str(band_tif_filepath)]
            run_subprocess(translate_to_tif, work_dir)

        band_tifpath_dict[int(band)] = band_tif_filepath
    return band_tifpath_dict   


def set_nodata_value(raster_path, nodata_value):
    """
    Set the NoData value for a given raster.

    Parameters:
    raster_path (str): The file path to the raster.
    nodata_value (numeric): The value to set as NoData.
    """
    # Open the raster file in update mode
    dataset = gdal.Open(str(raster_path), gdal.GA_Update)

    if dataset is None:
        raise FileNotFoundError(f"Raster file not found: {raster_path}")

    try:
        # Loop through all bands in the raster and set the NoData value
        for i in range(1, dataset.RasterCount + 1):
            band = dataset.GetRasterBand(i)
            band.SetNoDataValue(nodata_value)
            band.FlushCache()  # Write changes to disk
    finally:
        # Properly close the dataset to flush all data and avoid data corruption
        dataset = None

    log.debug(f"NoData value set to {nodata_value} for all bands in {raster_path}")
    
def read_raster_to_table(input_raster):
    dataset = gdal.Open(str(input_raster))
    bands = [dataset.GetRasterBand(i + 1).ReadAsArray() for i in range(dataset.RasterCount)]
    rows, cols = bands[0].shape
    data = np.array([band.flatten() for band in bands]).T
    return data, rows, cols


def extract_band(input_raster_path, band_number, output_raster_path = None):
    dataset = gdal.Open(str(input_raster_path))
    band = dataset.GetRasterBand(band_number)
    band_data = band.ReadAsArray()

    driver = gdal.GetDriverByName('GTiff')

    # Temporary output path
    temp_output_path = str(input_raster_path) + '.temp.tif'

    output_dataset = driver.Create(temp_output_path, band.XSize, band.YSize, 1, band.DataType)
    output_dataset.SetGeoTransform(dataset.GetGeoTransform())
    output_dataset.SetProjection(dataset.GetProjection())
    output_dataset.GetRasterBand(1).WriteArray(band_data)
    output_dataset.FlushCache()

    if output_raster_path is None:
        # Replace the original file
        os.remove(input_raster_path)
        os.rename(temp_output_path, input_raster_path)
    else:
        os.rename(temp_output_path, output_raster_path)


def sieve_multiclass_tif(input_raster, work_dir, output_raster = None, sieve_size = 7, diagonal_flag = False, background_value = 0):
    
    (xmin, ymax, RasterXSize, RasterYSize, pixel_width, projection, epsg, datatype, n_bands,
     imagery_extent_box) = read_raster_info(input_raster)
    # Get the data type of the band
    data_type_name = gdal.GetDataTypeName(datatype)
    # Now, get the constant from gdalconst
    data_type = getattr(gdalconst, 'GDT_' + data_type_name)
        
    rasterarray = raster2array(input_raster)
    rasterarray[np.isnan(rasterarray)] = background_value
    array_mask = np.zeros_like(rasterarray)
    array_mask[rasterarray != background_value] = 1
    array_mask[rasterarray == background_value] = 2
    array_mask_path = work_dir.joinpath(f"{input_raster.stem}_mask.tif")
    array_mask_sieved_path = work_dir.joinpath(f"{input_raster.stem}_sieved_mask.tif")
    save_raster_template(input_raster, array_mask_path, array_mask, data_type=GDT_Int16, nodata_value=0)

    # Use gdal_sieve.py to remove small patches based on the temporary mask
    create_sieved_raster(array_mask_path, array_mask_sieved_path, sieve=sieve_size, diagonal_flag=diagonal_flag)

    mask_array = raster2array(str(array_mask_sieved_path))
    rasterarray[mask_array != 1] = 0
    save_raster_template(str(input_raster), str(output_raster), rasterarray, data_type, nodata_value=background_value)
    log.debug(f"saved sieved raster to {str(output_raster)}")


def burn_first_digit(input_raster, output_raster):
    # Open the existing raster
    dataset = gdal.Open(str(input_raster), gdal.GA_ReadOnly)

    # Get raster band
    band = dataset.GetRasterBand(1)

    # Read raster data as array
    data = band.ReadAsArray()

    # Nodata value
    nodata = KULT_NO_DATAVALUE

    # Process the data to take only the first digit
    new_data = np.where(data == nodata, nodata, data // 10)

    # Create a new raster
    driver = gdal.GetDriverByName('GTiff')
    out_dataset = driver.Create(str(output_raster), dataset.RasterXSize, dataset.RasterYSize, 1, band.DataType)
    out_band = out_dataset.GetRasterBand(1)

    # Set geotransform and projection from the original data
    out_dataset.SetGeoTransform(dataset.GetGeoTransform())
    out_dataset.SetProjection(dataset.GetProjection())

    # Write the processed data to the new raster
    out_band.WriteArray(new_data)
    out_band.SetNoDataValue(nodata)
    out_band.FlushCache()

    # Close datasets
    out_dataset = None
    dataset = None

def vectorize_raster(raster_path, work_dir, field_name="class"):
    output_shp_path = Path(work_dir).joinpath(f"{raster_path.stem}.shp")
    output_gpkg_path = Path(work_dir).joinpath(f"{raster_path.stem}.gpkg")

    # Step 1: Use gdal_polygonize.py to convert raster to polygons (Shapefile)
    cmd_gdal = ['gdal_polygonize.py',
                str(raster_path),
                '-f', 'ESRI Shapefile',
                str(output_shp_path),
                'output_polygons',
                field_name]

    cmd_output = subprocess.run(cmd_gdal, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    log.debug("exit code {} --> {}".format(cmd_output.returncode, cmd_gdal))

    # Step 2: Use ogr2ogr to convert Shapefile to GeoPackage
    cmd_gdal = ['ogr2ogr',
                '-f', 'GPKG',
                str(output_gpkg_path),
                str(output_shp_path)]
    cmd_output = subprocess.run(cmd_gdal, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    log.debug("exit code {} --> {}".format(cmd_output.returncode, cmd_gdal))
    return output_gpkg_path



import logging
from datetime import datetime
from os.path import exists
import csv
import glob
import ogr
import osr
import os

from tondortools.tool import satellite_band

log = logging.getLogger(__name__)

def yearmonth_parse(yearmonth):
    start_final_date = [ym.strip() for ym in yearmonth.split("-")]
    start_date = datetime.strptime(start_final_date[0],"%Y%m%d")
    final_date   = datetime.strptime(start_final_date[1],"%Y%m%d")
    start_date_basename = start_date.strftime("%Y%m%d")
    final_date_basename = final_date.strftime("%Y%m%d")
    return yearmonth, start_date, final_date, start_date_basename, final_date_basename

def satellite_parse(satellites):
    satellite_masterlist = ['sentinel-2', 'landsat-4', 'landsat-5', 'landsat-7', 'landsat-8','landsat-9']
    satellites_list = [sat.strip() for sat in satellites.split(",")]

    if not set(satellites_list) <= set(satellite_masterlist):
       log.error("Satellite not in master list, check naming convention")

    basename_sat = ""
    for satellite in satellites_list:
        basename_sat = basename_sat + satellite[0] + satellite[-1]
    return satellites, satellites_list, basename_sat

def band_parse(bands, satellites, satellite_bandinstance, pixel_size):
    band_masterlist = ["coastal_aerosol", "vis-b", "vis-g", "vis-r", "nir", "swir1", "swir2", "pan", "cirrus", "tir-10900", "tir-12000", "tir-11450", "re-705", "re-740", "re-781"]
    bands_list = [band.strip() for band in bands.split(",")]

    if not set(bands_list) <= set(band_masterlist):
       log.error("Band not in master list, check naming convention")
    basename_band = ""
    for band in band_masterlist:
        if band in bands:
            basename_band = basename_band + "1"
            satellite_bandinstance.satellite_band_check(band, satellites, pixel_size)
        else:
            basename_band = basename_band + "0"
    return bands, bands_list, basename_band

def resampling_parse(resampling):
    resampling_masterlist = ['average', 'near', 'bilinear', 'cubic']
    if resampling not in resampling_masterlist:
       log.error("Resampling not in master list, check naming convention")
    return resampling

def statistic_parse(statistic, targetdates):
    statistic_masterlist = ["wavg", "spline"]
    if not statistic in statistic_masterlist:
       log.error("Statistic not in master list, check naming convention")

    targetdates_list = [tardat.strip() for tardat in targetdates.split(",")]
    return statistic, targetdates_list, targetdates

def pixel_parse(pixel_size):
    pixel_size_str = str(pixel_size)
    pixel_size_int = int(pixel_size)
    pixel_size_basename = 'PI' + pixel_size_str
    return pixel_size_int, pixel_size_str, pixel_size_basename

def maxcloudcover_parse(max_cloud_cover):
    max_cloud_cover_str = str(max_cloud_cover)
    max_cloud_cover_int = int(max_cloud_cover)
    max_cloud_cover_basename = 'MAXCLOUD' + max_cloud_cover_str
    return max_cloud_cover_int, max_cloud_cover_str, max_cloud_cover_basename

def trim_parse(trim_low, trim_high):
    trim_low_str  = str(trim_low)
    trim_high_str = str(trim_high)
    trim_low_int  = int(trim_low)
    trim_high_int = int(trim_high)
    trim_basename = 'TRIM' + trim_low_str.zfill(3) + trim_high_str.zfill(3)
    return trim_low_int, trim_high_int, trim_low_str, trim_high_str, trim_basename

def nodata_parse(nodata_value):
    nodata_value_int = int(nodata_value)
    nodata_value_str = str(nodata_value)
    nodata_value_basename = 'NODATA' + nodata_value_str
    return nodata_value_int, nodata_value_str, nodata_value_basename

def unmasksummersnow_parse(unmask_snow, start_date, final_date):
    unmasksummersnow_masterlist = ['Q3','07','08','09']
    unmask_summer_snow_basename = "UNMASKSUMSNOW"

    startdateyear  = start_date.strftime("%Y")
    startdatemonth = start_date.strftime("%m")
    startdateday   = start_date.strftime("%d")

    finaldateyear = final_date.strftime("%Y")
    finaldatemonth = final_date.strftime("%m")
    finaldateday   = final_date.strftime("%d")

    start = datetime(day=int(startdateday), month=int(startdatemonth), year=int(startdateyear))
    end   = datetime(day=int(finaldateday), month=int(finaldatemonth), year=int(finaldateyear))

    startsum = datetime.strptime("01-07-"+startdateyear, "%d-%m-%Y")
    endsum   = datetime.strptime("01-10-"+finaldateyear, "%d-%m-%Y")
    if startsum <= start <= endsum and startsum <= end <= endsum and unmask_snow:
       log.debug("The time period is within summer time and unmask snow is: {}".format(unmask_snow))
       unmask_snow_timeperiod = "yes"
       unmask_summer_snow_basename = unmask_summer_snow_basename + '1'
    else:
       log.debug("The time period is within summer time and unmask snow is: {}".format(unmask_snow))
       unmask_snow_timeperiod = "no"
       unmask_summer_snow_basename = unmask_summer_snow_basename + '0'
    return unmask_snow, unmask_summer_snow_basename

def find_epsg_tilename(sitecode):
    if sitecode[0:3] == "UTM":
       epsg_no = sitecode[3:5]
       EPSG = int('326' + epsg_no)
    else:
       epsg_no = sitecode[0:2]
       EPSG = int('326' + epsg_no)
    return EPSG

def get_sitecode_from_aoiwkt_parse(aoi_wkt, t_srs):
    if aoi_wkt.endswith(".gpkg"):
       log.debug("aoi is specified as an gpkg file")
       aoi_ds = ogr.Open(aoi_wkt)
       aoi_lyr = aoi_ds.GetLayer()
       aoi_feat = aoi_lyr.GetNextFeature()
       aoi_geom = aoi_feat.GetGeometryRef()
       aoi_geom_wkt = aoi_geom.ExportToWkt()
       aoi_srs = aoi_lyr.GetSpatialRef()
       aoi_epsg = aoi_srs.GetAttrValue('AUTHORITY', 1)
       log.debug("Espg of input gpkg file: {}".format(aoi_epsg))

       aoi_geom = get_aoi_geometry(aoi_wkt, int(aoi_epsg))
       if aoi_geom["status"] == "ERROR":
          logger.error(aoi_geom["message"])
       else:
          aoi_geom = aoi_geom["output"]

       log.debug("aoi_geom: {}".format(aoi_geom))
    else:
       aoi_geom = aoi_wkt
    if True:
       compositing_grid = get_grid_file(t_srs)
       if compositing_grid["status"] == "OK":
          log.debug("grid geopackage:{}".format(compositing_grid["output"]))
          compositing_squares = get_composite_squares(aoi_geom, compositing_grid["output"])
    return compositing_squares


def get_composite_squares(aoi_geom, grid_geopackage):
    """
    Return list of dictionaries - compositing squares in target Spatial Reference System.
    :param aoi_geom:
    :param grid_geopackage:
    :return:
    """
    aoi_geom_inner = ogr.CreateGeometryFromWkt(aoi_geom).Buffer(-1)

    # output dictionary
    composite_squares = list()

    # get grid geopackage layer
    grid_ds = ogr.Open(grid_geopackage)
    grid_lyr = grid_ds.GetLayer()
    aoi_geom = ogr.CreateGeometryFromWkt(aoi_geom)
    grid_lyr.SetSpatialFilter(aoi_geom)
    feat = grid_lyr.GetNextFeature()

    while feat is not None:
        feat_geom = feat.GetGeometryRef().Buffer(-1)

        # create little inner buffer to exclude neighbour squares
        if feat_geom.Intersects(aoi_geom_inner):

            # feat = grid_lyr.GetNextFeature()
            composite_square = {"id": feat.GetField("id"),
                                "sitecode":feat.GetField("sitecode"),
                                "xmin": feat.GetField("xmin"),
                                "xmax": feat.GetField("xmax"),
                                "ymin": feat.GetField("ymin"),
                                "ymax": feat.GetField("ymax"),
                                "sen2_tiles": feat.GetField("sen2_tiles").split(","),
                                "ls_tiles": feat.GetField("ls_tiles").split(","),
                                "sen1_orbits": feat.GetField("sen1_orb").split(",")}
            composite_squares.append(composite_square)
        feat = grid_lyr.GetNextFeature()
    # Sort the results by id.
    composite_squares = sorted(composite_squares, key=lambda sq: sq["sitecode"])
    sitecodes = [composite_square["sitecode"] for composite_square in composite_squares]
    return sitecodes, composite_squares

def tile_aoi(geom_wkt):
    """
    Check if aoi is large enough to split into tiles (larger than 4 compositing tiles).
    :param geom_wkt: aoi geometry wkt
    :return: True in case of large aoi; False in case of small aoi
    """
    geom = ogr.CreateGeometryFromWkt(geom_wkt)
    area = geom.GetArea()
    if area > 9e8:
        return True
    else:
        return False

def get_grid_file(epsg_code):
    """
    Get absolute path to compositing grid file.
    :param epsg_code: EPSG code of target Spatial Reference System.
    :return: absolute path to compositing grid file.
    """
    gridfile = os.path.abspath(os.path.join("/tondor/tiles_gpkgfiles", "{0}_withsitecode.gpkg".format(epsg_code)))
    if os.path.isfile(gridfile):
        return {"status": "OK",
                "output": gridfile}
    else:
        return {"status": "ERROR",
                "message": "Target SRS specified with EPSG code: {0} either does not exist or is not currently supported.".format(epsg_code)}

def get_aoi_geometry(aoi_file, epsg_code):
    """
    Get Union geometry WKT from vector spatial data file.
    :param aoi_file: file path to vector spatial data file (e.g.: .shp, .geojson, .gpkg...)
    :param epsg_code: EPSG code of target SRS (e.g.: 5514 for S-JTSK, 4326 for WGS-84...)
    :return: WKT of union geometry
    """

    if not os.path.isfile(aoi_file):
        return {"status": "ERROR",
                "message": "AOI file does not exist."}

    ds = ogr.Open(aoi_file)
    if ds is None:
        return {"status": "ERROR",
                "message": "AOI file is not in supported format."}
    lyr = ds.GetLayer()
    s_srs = lyr.GetSpatialRef()
    t_srs = osr.SpatialReference()
    t_srs.ImportFromEPSG(epsg_code)
    transform = osr.CoordinateTransformation(s_srs, t_srs)

    if lyr.GetFeatureCount() == 1:

        feat = lyr.GetNextFeature()
        geom = feat.GetGeometryRef()
        geom.Transform(transform)
        return {"status": "OK",
                "output": geom.ExportToWkt()}
    else:
        feat = lyr.GetNextFeature()
        geom = feat.GetGeometryRef()
        geom.Transform(transform)
        geom = geom.Clone()
        feat = lyr.GetNextFeature()

        while feat is not None:
            geom_next = feat.GetGeometryRef()
            geom_next.Transform(transform)
            geom = geom.Union(geom_next)
            feat = lyr.GetNextFeature()
        return {"status": "OK",
                "output": geom.ExportToWkt()}

    '''
       aoi_geom_processing_tiles_srs = ogr.CreateGeometryFromWkt(aoi_geom_wkt)
    elif aoi_wkt.lower().startswith("POLYGON".lower()) or aoi_wkt.lower().startswith("MULTIPOLYGON".lower()):
       log.debug("Extent specified")
       aoi_geom_processing_tiles_srs = ogr.CreateGeometryFromWkt(aoi_wkt)
    else:
       log.error("check aoi wkt parameter")

    processing_tiles_filepaths = glob.glob("/mnt/tondor.nfs/tiles_gpkgfiles/*.gpkg")

    for processing_tiles_filepath in processing_tiles_filepaths:
        # get processing tiles features
        processing_tiles_ds = ogr.Open(processing_tiles_filepath)
        processing_tiles_lyr = processing_tiles_ds.GetLayer()
        processing_tiles_srs = processing_tiles_lyr.GetSpatialRef()
        processing_tiles_epsg = processing_tiles_srs.GetAttrValue('AUTHORITY', 1)

        # transform AOI SRS to processing tiles's SRS if needed
        if aoi_epsg != processing_tiles_epsg:
            transform = osr.CoordinateTransformation(aoi_srs, processing_tiles_srs)
            aoi_geom_processing_tiles_srs.Transform(transform)

        processing_tiles_lyr.SetSpatialFilter(aoi_geom_processing_tiles_srs)
        nr_of_processing_units = processing_tiles_lyr.GetFeatureCount()
        log.debug("No of processing units:{} for tiles filepath {}".format(nr_of_processing_units, processing_tiles_filepath))
        if nr_of_processing_units >0:
                # processing iteratively over processing tiles
                processing_feature = processing_tiles_lyr.GetNextFeature()
                processing_feature_nr = 1
                while processing_feature is not None:
                    processing_feature_id = processing_feature.GetField("TILE_ID")
                    print("Processing feature ID is: {}".format(processing_feature_id))
                    processing_feature = processing_tiles_lyr.GetNextFeature()
    '''

def sitecode_extent(sitecode,  epsg_code):
    gridfile = os.path.abspath(os.path.join("/tondor/tiles_gpkgfiles", "{0}_withsitecode.gpkg".format(epsg_code)))
    processing_tiles_ds = ogr.Open(gridfile)
    processing_tiles_lyr = processing_tiles_ds.GetLayer()
    processing_tiles_srs = processing_tiles_lyr.GetSpatialRef()
    processing_tiles_epsg = processing_tiles_srs.GetAttrValue('AUTHORITY', 1)

    for feature in processing_tiles_lyr:

        if feature['sitecode'] == sitecode:
           return feature.geometry().ExportToWkt()

def load_included_scenes(csvfile):
    scene_identifiers = []
    with open(csvfile) as cf:
        csv_reader = csv.reader(cf, delimiter=",")
        for row in csv_reader:
            if row == []:
                break
            scene_identifiers.append(row[-1].strip())
    return scene_identifiers

def find_landsat_sentinel_fromincludedscenes(scenes):
    landsatscenes_dict = {}
    landsat9_dict = {}
    landsat8_dict = {}
    landsat7_dict = {}
    landsat5_dict = {}
    sentinelscenes_list = []
    for scene in scenes:
        scene_namepart_list = [sat.strip() for sat in scene.split("_")]
        if scene_namepart_list[0] == 'LC09':
            if scene_namepart_list[2] not in landsat9_dict.keys():
                landsat9_dict[scene_namepart_list[2]] = [scene_namepart_list[3]]
            else:
                landsat9_dict[scene_namepart_list[2]].append(scene_namepart_list[3])
        elif scene_namepart_list[0] == 'LC08':
            if scene_namepart_list[2] not in landsat8_dict.keys():
                landsat8_dict[scene_namepart_list[2]] = [scene_namepart_list[3]]
            else:
                landsat8_dict[scene_namepart_list[2]].append(scene_namepart_list[3])
        elif scene_namepart_list[0] == 'LC07':
            if scene_namepart_list[2] not in landsat7_dict.keys():
                landsat7_dict[scene_namepart_list[2]] = [scene_namepart_list[3]]
            else:
                landsat7_dict[scene_namepart_list[2]].append(scene_namepart_list[3])
        elif scene_namepart_list[0] == 'LC05':
            if scene_namepart_list[2] not in landsat5_dict.keys():
                landsat5_dict[scene_namepart_list[2]] = [scene_namepart_list[3]]
            else:
                landsat5_dict[scene_namepart_list[2]].append(scene_namepart_list[3])

        elif scene.startswith('S2'):
            scene_namepart_list = [sat.strip() for sat in scene.split(".")]
            sentinelscenes_list.append(scene_namepart_list[0])
    landsatscenes_dict[9] = landsat9_dict
    landsatscenes_dict[8] = landsat8_dict
    landsatscenes_dict[7] = landsat7_dict
    landsatscenes_dict[5] = landsat5_dict

    return landsatscenes_dict, sentinelscenes_list

def included_sceneslist_parse(INCLUDED_SCENES_PATH):

    includedscenes_file_exists = exists(INCLUDED_SCENES_PATH)
    included_scenes = []
    included_landsatscenes_dict ={}
    included_sentinelscenes_list = []
    if includedscenes_file_exists:
        log.debug("Included scenes file exist in :{}".format(INCLUDED_SCENES_PATH))
        included_scenes = load_included_scenes(INCLUDED_SCENES_PATH)
        log.debug("Number of all included scenes are :{}".format(len(included_scenes)))
        included_landsatscenes_dict, included_sentinelscenes_list = find_landsat_sentinel_fromincludedscenes(included_scenes)

    if INCLUDED_SCENES_PATH != "" and len(included_scenes) == 0:
       input("Included scene path is one of the input. But the file is not found. Check whether the link is correct. Exiting")
       raise Exception("Incorrect included scene file")
       exit()

    return included_landsatscenes_dict, included_sentinelscenes_list, includedscenes_file_exists, included_scenes

def ignore_skip_missingL2A_parse(SKIP_MISSING_L2A_PATH):
    skip_missing_l2a_exists = exists(SKIP_MISSING_L2A_PATH)
    missing_l2a_list = []
    if skip_missing_l2a_exists:
       log.debug("Skipped missing l2a scenes file exist in :{}".format(SKIP_MISSING_L2A_PATH))
       missing_l2a_list = load_included_scenes(SKIP_MISSING_L2A)
    if SKIP_MISSING_L2A_PATH != "" and len(missing_l2a_list) == 0:
       input("Skipped missing l2a file list is one of the input. But the file is not found. Check whether the link is correct. Exiting")
       log.error("Incorrect skipping missing l2a file")
       exit()

    return skip_missing_l2a_exists, missing_l2a_list

def compile_basename_optcomposite(optcomposite_pattern_dict):

    optcomposite_basename = 'OPT_STARTDATE_FINALDATE_SITECODE_PIXELSIZE_SATELLITES_BANDS_RESAMPLING_UNMASKSUMMERSNOW_MAXCLOUD_TRIM_DATATYPE_OUTPUTNODATAVALUE_STATISTIC_*'
    optcomposite_master_pattern_list = ['SATELLITES', 'PIXEL_SIZE', 'BANDS', 'DATE_RANGE', 'SITECODE', 'STATISTICS', 'RESAMPLING', 'TARGET_DATES', 'OUTPUT_NODATA_VALUE',
                                   'MAX_CLOUD', 'TRIM_LOW', 'TRIM_HIGH', 'UNMASK_SUMMER_SNOW']

    optcomposite_basename_parts_list = optcomposite_basename.split('_')
    satellite_bandinstance  = satellite_band()

    for item in optcomposite_master_pattern_list:
        if item in optcomposite_pattern_dict.keys():

            if item == 'SATELLITES':
                satellites, satellites_list, basename_sat = satellite_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('SATELLITES', basename_sat)

            elif item == 'PIXEL_SIZE':
                pixel_size, pixel_size_str, pixel_size_basename = pixel_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('PIXELSIZE', pixel_size_basename)

            elif item == 'BANDS':
                all_binary = all(c in '01' for c in optcomposite_pattern_dict[item])
                if all_binary:
                    optcomposite_basename = optcomposite_basename.replace('BANDS', optcomposite_pattern_dict[item])
                    continue
                _, bands, basename_band = band_parse(optcomposite_pattern_dict[item], satellites, satellite_bandinstance, pixel_size)
                optcomposite_basename = optcomposite_basename.replace('BANDS',basename_band)

            elif item == 'STATISTICS':
                statistics = 'wavg'
                optcomposite_basename = optcomposite_basename.replace('STATISTIC',statistics)

            elif item == 'DATE_RANGE':
                yearmonth, start_date, final_date, start_date_basename, final_date_basename = yearmonth_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('STARTDATE',start_date_basename)
                optcomposite_basename = optcomposite_basename.replace('FINALDATE',final_date_basename)

            elif item == 'SITECODE':
                optcomposite_basename = optcomposite_basename.replace('SITECODE', optcomposite_pattern_dict[item])

            elif item == 'RESAMPLING':
                resampling = resampling_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('RESAMPLING', resampling)

            elif item == 'MAX_CLOUD':
                max_cloud_cover, max_cloud_cover_str, max_cloud_cover_basename = maxcloudcover_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('MAXCLOUD', max_cloud_cover_basename)

            elif item == 'UNMASK_SUMMER_SNOW':
                if optcomposite_pattern_dict[item] == 'yes': unmask_snow = True
                else: unmask_snow = False
                unmask_snow, unmask_summer_snow_basename = unmasksummersnow_parse(unmask_snow, start_date, final_date)
                optcomposite_basename = optcomposite_basename.replace('UNMASKSUMMERSNOW', unmask_summer_snow_basename)

    for optcomposite_basename_parts_listitem in optcomposite_basename_parts_list:
        if optcomposite_basename_parts_listitem == 'OPT':
            continue
        optcomposite_basename = optcomposite_basename.replace(optcomposite_basename_parts_listitem, '*')
    log.debug(optcomposite_basename)
    return optcomposite_basename

def compile_optcomposite_nodata_basename(optcomposite_pattern_dict, look_for_pattern=False):

    optcomposite_basename = 'OPT_STARTDATE_FINALDATE_SITECODE_PIXELSIZE_SATELLITES_BANDS_NODATA.txt'
    optcomposite_master_pattern_list = ['SATELLITES', 'PIXEL_SIZE', 'BANDS', 'DATE_RANGE', 'SITECODE']


    optcomposite_basename_parts_list = optcomposite_basename.split('_')
    satellite_bandinstance  = satellite_band()

    for item in optcomposite_master_pattern_list:
        if item in optcomposite_pattern_dict.keys():

            if item == 'SATELLITES':
                satellites, satellites_list, basename_sat = satellite_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('SATELLITES', basename_sat)

            elif item == 'PIXEL_SIZE':
                pixel_size, pixel_size_str, pixel_size_basename = pixel_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('PIXELSIZE', pixel_size_basename)

            elif item == 'BANDS':
                all_binary = all(c in '01' for c in optcomposite_pattern_dict[item])
                if all_binary:
                    optcomposite_basename = optcomposite_basename.replace('BANDS', optcomposite_pattern_dict[item])
                    continue
                _, bands, basename_band = band_parse(optcomposite_pattern_dict[item], satellites, satellite_bandinstance, pixel_size)
                optcomposite_basename = optcomposite_basename.replace('BANDS',basename_band)

            elif item == 'STATISTICS':
                statistics = 'wavg'
                optcomposite_basename = optcomposite_basename.replace('STATISTIC',statistics)

            elif item == 'DATE_RANGE':
                yearmonth, start_date, final_date, start_date_basename, final_date_basename = yearmonth_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('STARTDATE',start_date_basename)
                optcomposite_basename = optcomposite_basename.replace('FINALDATE',final_date_basename)

            elif item == 'SITECODE':
                optcomposite_basename = optcomposite_basename.replace('SITECODE', optcomposite_pattern_dict[item])

            elif item == 'RESAMPLING':
                resampling = resampling_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('RESAMPLING', resampling)

            elif item == 'MAX_CLOUD':
                max_cloud_cover, max_cloud_cover_str, max_cloud_cover_basename = maxcloudcover_parse(optcomposite_pattern_dict[item])
                optcomposite_basename = optcomposite_basename.replace('MAXCLOUD', max_cloud_cover_basename)

            elif item == 'UNMASK_SUMMER_SNOW':
                if optcomposite_pattern_dict[item] == 'yes': unmask_snow = True
                else: unmask_snow = False
                unmask_snow, unmask_summer_snow_basename = unmasksummersnow_parse(unmask_snow, start_date, final_date)
                optcomposite_basename = optcomposite_basename.replace('UNMASKSUMMERSNOW', unmask_summer_snow_basename)

    for optcomposite_basename_parts_listitem in optcomposite_basename_parts_list:
        if optcomposite_basename_parts_listitem == "OPT" or optcomposite_basename_parts_listitem == "NODATA.txt":
            continue
        optcomposite_basename = optcomposite_basename.replace(optcomposite_basename_parts_listitem, '*')
    log.debug(optcomposite_basename)
    if not look_for_pattern:
        if '*' in optcomposite_basename:
            raise Exception("provide correct dict to create nodata txt filename")
    return optcomposite_basename


import logging
import os
from time import gmtime

def get_console_handler():
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(fmt="{levelname} {message}", style="{"))
    return console_handler

def init_logging():
    logging.Formatter.converter = gmtime
    root_log = logging.getLogger()
    root_log.addHandler(get_console_handler())
    root_log.setLevel(logging.WARNING)
    logging.getLogger('pandas').setLevel(logging.WARNING)
    logging.getLogger('geopandas').setLevel(logging.WARNING)
    logging.getLogger('rasterio').setLevel(logging.WARNING)
    logging.getLogger('PROJ').setLevel(logging.CRITICAL)
    
    os.environ['PROJ_DEBUG'] = '0'
